package main

import (
	"strings"

	"git.sr.ht/~tslocum/gmenu/pkg/gmenu"
	"github.com/gdamore/tcell"
	"github.com/rivo/tview"
)

var (
	app            *tview.Application
	inputView      *tview.InputField
	optionsList    *OptionsList
	appDetailsView *tview.TextView

	closedGUI bool
)

type OptionsList struct {
	*tview.TextView
	options  []string
	origin   int
	selected int
	shown    int
}

func NewOptionsList(options []string) *OptionsList {
	opts := OptionsList{
		TextView: tview.NewTextView(),
		options:  options,
	}

	opts.TextView = opts.SetDynamicColors(true).SetWordWrap(false)

	return &opts
}

func (r *OptionsList) Draw(screen tcell.Screen) {
	_, height := screen.Size()

	var b strings.Builder
	r.shown = 0
	for i, option := range r.options {
		if i < r.origin || i-r.origin >= height-1 {
			continue
		}

		if i == r.selected {
			b.WriteString(`[::r]`)
		}
		if i-r.origin < height-2 {
			b.WriteString(option + "\n")
		} else {
			b.WriteString(option)
		}
		if i == r.selected {
			b.WriteString(`[-:-:-]`)
		}

		r.shown++
	}

	r.TextView.SetText(b.String()).Highlight("gmenu").Draw(screen)
}

func initGUI() (*tview.Application, error) {
	app = tview.NewApplication()
	inputView = tview.NewInputField().
		SetLabel("").
		SetFieldWidth(0).
		SetFieldBackgroundColor(tcell.ColorDefault).
		SetFieldTextColor(tcell.ColorDefault).
		SetChangedFunc(func(text string) {
			gmenu.SetInput(text)
		})

	optionsList = NewOptionsList(nil)

	grid := tview.NewGrid().
		SetBorders(false).
		SetRows(1, -1)

	appDetailsView = tview.NewTextView().
		SetTextAlign(tview.AlignLeft).
		SetWrap(true).
		SetWordWrap(true)

	if config.HideAppDetails {
		grid = grid.SetColumns(-1).
			AddItem(inputView, 0, 0, 1, 1, 0, 0, true).
			AddItem(optionsList, 1, 0, 1, 1, 0, 0, false)
	} else {
		grid = grid.SetColumns(-1, -1).
			AddItem(inputView, 0, 0, 1, 2, 0, 0, true).
			AddItem(optionsList, 1, 0, 1, 1, 0, 0, false).
			AddItem(appDetailsView, 1, 1, 1, 1, 0, 0, false)
	}

	app = app.SetInputCapture(func(event *tcell.EventKey) *tcell.EventKey {
		if event.Key() == tcell.KeyUp {
			if optionsList.origin > 0 && optionsList.selected == optionsList.origin {
				optionsList.origin--
			}
			if optionsList.selected > 0 {
				optionsList.selected--
			}
			event = nil
			updateEntryInfo()
			app.Draw()
		} else if event.Key() == tcell.KeyDown {
			if optionsList.selected < len(optionsList.options)-1 {
				optionsList.selected++
				if optionsList.selected > optionsList.origin+optionsList.shown-1 {
					optionsList.origin++
				}
			}
			event = nil
			updateEntryInfo()
			app.Draw()
		} else if event.Key() == tcell.KeyPgUp {
			if optionsList.origin == 0 {
				optionsList.selected = 0

				updateEntryInfo()
				app.Draw()
				return nil
			}

			optionsList.origin -= optionsList.shown - 2
			if optionsList.origin < 0 {
				optionsList.origin = 0
			}
			optionsList.selected = optionsList.origin

			updateEntryInfo()
			app.Draw()
			return nil
		} else if event.Key() == tcell.KeyPgDn {
			numEntries := len(gmenu.FilteredEntries)

			if optionsList.origin >= numEntries-optionsList.shown {
				optionsList.selected = numEntries - 1

				updateEntryInfo()
				app.Draw()
				return nil
			}

			optionsList.origin += optionsList.shown - 2
			if optionsList.origin > numEntries-optionsList.shown {
				optionsList.origin = numEntries - optionsList.shown
			}
			optionsList.selected = optionsList.origin

			updateEntryInfo()
			app.Draw()
			return nil
		} else if event.Key() == tcell.KeyEnter {
			err := listSelect()
			if err != nil {
				panic(err)
			}
		} else if event.Key() == tcell.KeyEscape {
			done <- true
		}
		return event
	})

	app = app.SetRoot(grid, true)

	go gmenu.HandleInput(updateEntries)
	gmenu.SetInput("")
	return app, nil
}

func closeGUI() {
	if closedGUI {
		return
	}
	closedGUI = true

	app.Stop()
}
