package gmenu

import (
	"flag"
	"fmt"
	"log"
	"sort"
	"strings"

	"git.sr.ht/~tslocum/desktop"
	"github.com/lithammer/fuzzysearch/fuzzy"
	"github.com/pkg/errors"
)

var (
	Entries []*desktop.Entry
	Names   []string

	FilteredEntries []*ListEntry

	inputBuffer  = make(chan string, 3)
	input        string
	inputLower string
	inputFlushed = make(chan bool)
)

type ListEntry struct {
	*desktop.Entry

	Label string
}

type InputUpdateHandler func(input string)

func SharedInit(c *Config) {
	log.SetFlags(0)

	flag.StringVar(&c.DataDirs, "data-dirs", "", "application data directories (default: $XDG_DATA_DIRS)")
	flag.BoolVar(&c.HideGenericNames, "no-generic", false, "hide application generic names")
	flag.BoolVar(&c.HideAppDetails, "no-details", false, "hide application details")
	flag.StringVar(&c.terminalCommand, "terminal", "", "terminal command")
	flag.StringVar(&c.browserCommand, "browser", "", "browser command")
}

func HandleInput(u InputUpdateHandler) {
	var (
		in string
		ok bool
	)

inputLoop:
	for {
		select {
		case in, ok = <-inputBuffer:
			if !ok {
				break inputLoop
			}

			input = in
			inputLower = strings.ToLower(in)
			
			u(input)
		}
	}

	inputFlushed <- true
}

func LoadEntries(c *Config) {
	var err error
	Entries, Names, err = DesktopEntries(c)
	if err != nil {
		log.Fatal(err)
	}
}

func SetInput(i string) {
	inputBuffer <- i
}

func CloseInput() {
	close(inputBuffer)
	<-inputFlushed
}

func MatchEntry(i int) bool {
	if i == -1 {
		return true
	}

	if inputLower == "" {
		return true
	}

	return fuzzy.MatchFold(inputLower, Names[i])
}

func FilterEntries() {
	FilteredEntries = nil
	if input == "" {
		for i, l := range Names {
			FilteredEntries = append(FilteredEntries, &ListEntry{Label: l, Entry: Entries[i]})
		}

		sort.Slice(FilteredEntries, SortEmpty)
	} else {
		matches := fuzzy.RankFindFold(inputLower, Names)
		sort.Sort(matches)

		for _, match := range matches {
			FilteredEntries = append(FilteredEntries, &ListEntry{Label: Names[match.OriginalIndex], Entry: Entries[match.OriginalIndex]})
		}

		sort.Slice(FilteredEntries, SortFiltered)
	}
}

func DesktopEntries(c *Config) ([]*desktop.Entry, []string, error) {
	var dirs []string
	if c.DataDirs != "" {
		dirs = strings.Split(c.DataDirs, ":")
	} else {
		dirs = desktop.DataDirs()
	}

	allEntries, err := desktop.Scan(dirs)
	if err != nil {
		return nil, nil, err
	}

	var (
		desktopEntries []*desktop.Entry
		desktopNames   []string
	)

	for _, entries := range allEntries {
		for _, entry := range entries {
			switch entry.Type {
			case desktop.Application:
				if entry.Exec == "" {
					continue
				}
			case desktop.Link:
				if entry.URL == "" {
					continue
				}
			default:
				continue // Unsupported entry type
			}

			if entry.Name != "" {
				desktopEntries = append(desktopEntries, entry)
				desktopNames = append(desktopNames, entry.Name)
			}
			if !c.HideGenericNames && entry.GenericName != "" {
				desktopEntries = append(desktopEntries, entry)
				desktopNames = append(desktopNames, entry.GenericName)
			}
		}
	}

	return desktopEntries, desktopNames, nil
}

func Sort(i, j int) bool {
	if input == "" {
		return SortEmpty(i, j)
	}

	return SortFiltered(i, j)
}

func SortEmpty(i, j int) bool {
	ilower := strings.ToLower(FilteredEntries[i].Label)
	jlower := strings.ToLower(FilteredEntries[j].Label)

	if FilteredEntries[i].Entry == nil && FilteredEntries[j].Entry != nil {
		return true
	} else if ilower != jlower {
		return ilower < jlower
	} else {
		return i < j
	}
}

func SortFiltered(i, j int) bool {
	ilower := strings.ToLower(FilteredEntries[i].Label)
	if FilteredEntries[i].Entry == nil {
		ilower = ""
	}

	jlower := strings.ToLower(FilteredEntries[j].Label)
	if FilteredEntries[j].Entry == nil {
		jlower = ""
	}

	ipre := strings.HasPrefix(ilower, inputLower)
	jpre := strings.HasPrefix(jlower, inputLower)

	icon := strings.Contains(ilower, inputLower)
	jcon := strings.Contains(jlower, inputLower)

	imatch := fuzzy.MatchFold(inputLower, ilower)
	jmatch := fuzzy.MatchFold(inputLower, jlower)

	if ipre != jpre {
		return ipre && !jpre
	} else if icon != jcon {
		return icon && !jcon
	} else if imatch != jmatch {
		return imatch && !jmatch
	} else if (FilteredEntries[i].Entry == nil) != (FilteredEntries[j].Entry == nil) {
		return FilteredEntries[i].Entry == nil
	} else if ilower != jlower {
		return ilower < jlower
	} else {
		return i < j
	}
}

func Run(config *Config, execute string, path string, runInTerminal bool, waitUntilFinished bool) error {
	execute = strings.TrimSpace(execute)

	fmt.Println(execute)

	runScript, err := desktop.RunScript(execute)
	if err != nil {
		return errors.Wrap(err, "failed to create run script")
	}

	return run(config, runScript, path, waitUntilFinished, runInTerminal)
}
